// SoC (Sphere of Confusion) March by eiffie, to be used with nullTrace.cl
// License Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License.

// these should be parameters
#define focalDistance 2.0f
#define aperture 0.01f
#define fudgeFactor 1.0f
#define shadowCone 0.5f
#define pixelSize 0.002f

float CircleOfConfusion(const float t){//calculates the radius of the circle of confusion at length t
	return max(fabs(focalDistance-t)*aperture,pixelSize*(1.0f+t));
}

float linstep(const float a, const float b, const float t){
	return min(max((t-a)/(b-a),0.0f),1.0f);
}

float rand(float2 co){// implementation found at: lumina.sourceforge.net/Tutorials/Noise.html
	float f=fabs(sin(dot(co*917.123,(float2)(12.9898,78.233))) * 43758.5453);
	return f-floor(f);
}

float FuzzyShadow(const float3 roo, const float3 rd, const float coneGrad, const float rCoC, __constant float* a_Parameters){
	float t=0.0,d=1.0,s=1.0;
	float3 ro=roo+rd*0.01;
	for(int i=0;i<12;i++){
		if(s<0.1)break;
		float r=rCoC+t*coneGrad;//radius of cone
		d=DistanceEstimation(ro+rd*t, a_Parameters)+r*0.5;
		s*=linstep(-r,r,d);
		t+=fabs(d)*(0.8+0.2*rand(ro.xy+rd.xy+(float2)(i,i)));
	}
	return min(max(s*0.75+0.25,0.0),1.0);
}


float3 GetNormal(const float3 a_Position, const float delta, __constant float* a_Parameters)
{
	float3 n;
	float3 offset = (float3)(delta, 0.0f, 0.0f);
	
	n.x = DistanceEstimation(a_Position + offset.xyz, a_Parameters) - DistanceEstimation(a_Position - offset.xyz, a_Parameters);
	n.y = DistanceEstimation(a_Position + offset.yxz, a_Parameters) - DistanceEstimation(a_Position - offset.yxz, a_Parameters);
	n.z = DistanceEstimation(a_Position + offset.yzx, a_Parameters) - DistanceEstimation(a_Position - offset.yzx, a_Parameters);
	
	return normalize(n);
}
float3 Reflect(const float3 a_Ray, const float3 a_Normal)
{
	//R - T(2) * N * Dot(R, N);

	return a_Ray - ((float3)(2.0f)) * a_Normal * dot(a_Ray, a_Normal);
}

float4 Shade(const float3 a_Position, const float3 a_Ray, const float2 a_Steps, __constant float* a_Parameters)
{
	float3 ro=a_Position;
	float3 rd=a_Ray;
	float3 L=normalize((float3)(0.5,0.5,-0.5));
	float4 col=(float4)(0.0,0.0,0.0,0.0);
	float t=0.0;
	for(int i=0;i<a_Steps.y;i++){
		if(col.w>0.9)break;
		float rCoC=CircleOfConfusion(t);//calc the radius of CoC
		float d=DistanceEstimation(ro, a_Parameters)+0.25*rCoC;
		if(d<rCoC){//if we are inside add its contribution
			float3 p=ro-rd*(rCoC*0.25-d);//back up to border of CoC
			float3 N=GetNormal(p, rCoC*0.1, a_Parameters);
			//if(N!=N)N=-rd;
			float k=max(0.333,dot(N,-rd));
			float3 scol=(float3)(0.7,0.6,0.5)*(0.2+0.4*(1.0+dot(N,L)));
			scol+=(float3)(0.5,0.25,0.0)*pow(max(0.0,dot(Reflect(rd,N),L)),32.0);
			scol*=FuzzyShadow(p,L,shadowCone,rCoC,a_Parameters);
			float alpha=fudgeFactor*k*(1.0-col.w)*linstep(-rCoC,rCoC,-d);//calculate the mix like cloud density
			col+=(float4)(scol*alpha,alpha);//blend in the new color
		}
		d=fudgeFactor*max(d,pixelSize)*(0.8+0.2*rand(ro.xy+rd.xy+(float2)(i,i)));//add in noise to reduce banding and create fuzz
		ro+=d*rd;//march
		t+=d;
	}//mix in background color
	float3 scol=(float3)(0.5,0.5,0.5)+rd*0.1;
	col.xyz+=scol*(1.0-min(max(col.w,0.0),1.0));
	
	return (float4)(col.xyz, 1.0f);
}