// simple phong shader with hard shadows
// the color and light position should be in parameters
// use with the default tracer

float3 GetNormal(const float3 a_Position, __constant float* a_Parameters)
{
	float3 n;
	float3 offset = (float3)(0.0001f, 0.0f, 0.0f);
	
	n.x = DistanceEstimation(a_Position + offset.xyz, a_Parameters) - DistanceEstimation(a_Position - offset.xyz, a_Parameters);
	n.y = DistanceEstimation(a_Position + offset.yxz, a_Parameters) - DistanceEstimation(a_Position - offset.yxz, a_Parameters);
	n.z = DistanceEstimation(a_Position + offset.yzx, a_Parameters) - DistanceEstimation(a_Position - offset.yzx, a_Parameters);
	
	return normalize(n);
}
float3 Reflect(const float3 a_Ray, const float3 a_Normal)
{
	//R - T(2) * N * Dot(R, N);

	return a_Ray - ((float3)(2.0f)) * a_Normal * dot(a_Ray, a_Normal);
}

float4 Shade(const float3 a_Position, const float3 a_Ray, const float2 a_Steps, __constant float* a_Parameters)
{
	float3 C=(float3)(0.7,0.6,0.5);
	float3 N=GetNormal(a_Position, a_Parameters);
	float3 L=normalize((float3)(0.5,0.5,-0.5));
	float3 R=Reflect(a_Ray,N);

	float3 position = a_Position + L * EPSILON * 5.0f;
	float4 pos = Trace(position, L, a_Steps.y, a_Parameters);
	float shadow = 1.0;
	if(pos.w>0.0)shadow=0.0;

	C=C*(0.2+shadow*0.8*max(0.0,dot(N,L)));
	C=C+shadow*0.5*pow(max(0.0,dot(R,L)),32.0);
	return (float4)(C, 1.0f);
}